set cell_name [lindex $argv 0]
set pblock_name [lindex $argv 1]

set cell [get_cell $cell_name]

report_utilization -pblocks $pblock_name -file pblock_util.rpt
report_timing_summary -cells $cell -warn_on_violation -file timing.rpt
