
{% for platform in platforms %}
{{hw}}_{{platform}} = get_platform('{{hw}}-{{platform}}')
{% endfor %}

{% for platform in platforms %}
alias('{{hw}}-{{platform}}',
    publish(BarFile('{{name}}-{{hw}}-{{platform}}.bar',
        create_{{name}}({{hw}}_{{platform}}),
    )),
)
{% endfor %}

alias('{{hw}}', [
{% for platform in platforms %}
    '{{hw}}-{{platform}}',
{% endfor %}
])

