alias('slot-vhdl', source('vhdl/*.vhd'))
alias('shell-zynqmp-tcl', source('tcl/shell-zynqmp.tcl'))
alias('shell-zynq-tcl', source('tcl/shell-zynq.tcl'))
alias('slot-synth-tcl', source('tcl/synth.tcl'))
alias('slot-route-tcl', source('tcl/route.tcl'))
alias('slot-reports-tcl', source('tcl/reports.tcl'))
alias('slot-constraints', source('constraints/slot_default.xdc'))
alias('slot-bitstream-tcl', source('tcl/bitstream.tcl'))
alias('slot-python', source('python/*.py'))
alias('sim-python', source('sim/*.py'))
