set synth_dcp [lindex $argv 0]
set cell_name [lindex $argv 1]
set dcp_out [lindex $argv 2]

set cell [get_cell $cell_name]

read_checkpoint -cell $cell $synth_dcp

read_xdc [glob constraints/*.xdc]

opt_design -directive Explore
place_design -directive Explore -no_bufg_opt
phys_opt_design -directive AggressiveExplore
route_design -directive AggressiveExplore
phys_opt_design -directive AggressiveExplore

write_checkpoint -force $dcp_out
