set cell_name [lindex $argv 0]
set bit [lindex $argv 1]

set cell [get_cell $cell_name]

set_property BITSTREAM.GENERAL.COMPRESS TRUE [current_design]
#set_property BITSTREAM.ENCRYPTION.ENCRYPT YES [get_designs impl_1]
#set_property BITSTREAM.ENCRYPTION.ENCRYPTKEYSELECT EFUSE [get_designs impl_1]
#BITSTREAM.ENCRYPTION.KEYFILE

write_bitstream -cell $cell -force -file $bit
