# zynq devices we need to lock the routing of the entire platform

set cell [get_cell [lindex $argv 0]]
set dcp [lindex $argv 1]

update_design -cell $cell -black_box
lock_design -level routing

write_checkpoint -force $dcp
